package meerkat;

import meerkat.bulletinboard.BulletinBoardClient;
import meerkat.crypto.DigitalSignature;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.BulletinBoardAPI.BulletinBoardMessage;
import meerkat.protobuf.BulletinBoardAPI.MessageFilter;
import meerkat.protobuf.BulletinBoardAPI.MessageFilterList;
import meerkat.protobuf.BulletinBoardAPI.UnsignedBulletinBoardMessage;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.RegistryTags;

import java.security.SignatureException;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class SimpleRegistry  implements VoterRegistry{

    protected DigitalSignature signatory;
    protected BulletinBoardClient bulletinBoardClient ;

    public SimpleRegistry(DigitalSignature signatory, BulletinBoardClient communicator) {
        this.signatory = signatory;
        this.bulletinBoardClient  = communicator;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and
     *         UnsignedBulletinBoardMessage that contains the basic message
     *
     * @param basicMessage BasicMessage
     * @return BulletinBoardMessage
     */
    private BulletinBoardMessage CreateBulletinBoardMessage(UnsignedBulletinBoardMessage basicMessage) {
        try {
            BulletinBoardMessage.Builder bulletinBoardMessage =
                    BulletinBoardMessage.newBuilder();

            signatory.updateContent(basicMessage);

            Crypto.Signature signature = signatory.sign();
            bulletinBoardMessage.addSig(signature);

            bulletinBoardMessage.setMsg(basicMessage);

            return bulletinBoardMessage.build();
        } catch (SignatureException e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return MessageFilterList.
     */
    public MessageFilterList GetRelevantMessagesFilters(List<String> tags) {
        MessageFilterList.Builder filters = MessageFilterList.newBuilder();

        if (tags == null){
            return filters.build();
        }

        for (String tag : tags) {
            MessageFilter.Builder filter =
                    MessageFilter.newBuilder().setTag(tag)
                            .setType(BulletinBoardAPI.FilterType.TAG);

            filters.addFilter(filter);
        }
        return filters.build();
    }

    /**
     * Tries to post basicMessage and return true if its been successfully posted
     *
     * @param Message the massage to post
     * @return true when the post was successful else false
     */
    private void SafePost(BulletinBoardMessage Message, BulletinBoardClient.ClientCallback callback) {
        bulletinBoardClient .postMessage(Message, callback);
    }

    public void AddVoter(RegistryMessages.VoterInfo voterInfo, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + voterInfo.getId().getId())
                        .addTag(RegistryTags.VOTER_ENTRY_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        basicMessage.setData(voterInfo.getInfoBytes());

        SafePost(CreateBulletinBoardMessage(basicMessage.build()), new BooleanHandler(callback));
    }

    public void AddToGroup(RegistryMessages.VoterGroup voterGroup, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterGroup.getVoterId().getId())
                        .addTag(RegistryTags.GROUP_ID_TAG + voterGroup.getGroupId().getId())
                        .addTag(RegistryTags.GROUP_ACTION_TAG .toString() + RegistryTags.ADD_TO_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        SafePost(CreateBulletinBoardMessage(basicMessage.build()), new BooleanHandler(callback));
    }

    public void SetVoted(RegistryMessages.VoterID voterId, RegistryCallBack callback) {
        UnsignedBulletinBoardMessage.Builder basicMessage =
                UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + voterId.getId())
                        .addTag(RegistryTags.VOTE_ACTION_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + AccurateTimestamp.GetCurrentTimestampString());

        SafePost(CreateBulletinBoardMessage(basicMessage.build()), new BooleanHandler(callback));
    }

    public void GetGroups(RegistryMessages.GroupID groupID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>() {{
           add(RegistryTags.GROUP_ID_TAG + groupID.getId());
        }};
        bulletinBoardClient.readMessages(GetRelevantMessagesFilters(GroupsActionsTags),
                new RelevantDataCallBack(callback));
    }

    public void GetPersonIDDetails(RegistryMessages.VoterID voterID, RegistryCallBack callback) {
        List<String> GroupsActionsTags = new ArrayList<String>() {{
            add(RegistryTags.ID_TAG + voterID.getId());
            add(RegistryTags.VOTER_ENTRY_TAG.toString());
        }};
        bulletinBoardClient.readMessages(GetRelevantMessagesFilters(GroupsActionsTags),
                new RelevantDataCallBack(callback));
    }
}
